@@ -4,7 +4,7 @@ package LWPx::UserAgent::Cached;
 
 use Modern::Perl '2011';    ## no critic (Modules::ProhibitUseQuotedVersion)
 
-our $VERSION = '0.004';     # VERSION
+our $VERSION = '0.006';     # VERSION
 use utf8;
 
 ## no critic (Bangs::ProhibitCommentedOutCode)
@@ -29,6 +29,13 @@ use utf8;
 #pod instead it is just a direct subclass of
 #pod L<LWP::UserAgent|LWP::UserAgent>.
 #pod
+#pod =head2 HTTP/1.1 cache operation
+#pod
+#pod Full HTTP/1.1 cache compliance is a work in progress. As of version 0.006 we
+#pod have limited support for HTTP/1.1 C<ETag>/C<If-None-Match> headers, as well as
+#pod C<no-cache> and C<no-store> C<Cache-Control> directives (both on request and
+#pod response) and the C<Pragma: no-cache> request header.
+#pod
 #pod =head1 SEE ALSO
 #pod
 #pod =over
@@ -46,14 +53,9 @@ use utf8;
 #pod =cut
 
 use CHI;
-use HTTP::Status qw(HTTP_OK HTTP_MOVED_PERMANENTLY);
-
-# work around RT#43310
-## no critic (Subroutines::ProhibitCallsToUndeclaredSubs)
+use HTTP::Status qw(HTTP_OK HTTP_MOVED_PERMANENTLY HTTP_NOT_MODIFIED);
 use List::Util 1.33 'any';
-use Sereal qw(sereal_encode_with_object sereal_decode_with_object);
 use Moo 1.004005;
-use Sub::Quote 1.005000 'qsub';
 use Types::Standard qw(Bool HasMethods HashRef InstanceOf Maybe);
 use namespace::clean;
 extends 'LWP::UserAgent';
@@ -73,9 +75,10 @@ has cache => (
     isa     => HasMethods [qw(get set)],
     default => sub {
         CHI->new(
-            driver    => 'RawMemory',
-            datastore => $_[0]->_cache_datastore,
-            namespace => __PACKAGE__,
+            serializer => 'Sereal',
+            driver     => 'RawMemory',
+            datastore  => $_[0]->_cache_datastore,
+            namespace  => __PACKAGE__,
         );
     },
 );
@@ -141,14 +144,6 @@ has ref_in_cache_key => ( is => 'rw', isa => Bool, default => 0 );
 
 has positive_cache => ( is => 'rw', isa => Bool, default => 1 );
 
-for (qw(Encoder Decoder)) {
-    has "_\l$_" => (
-        is      => 'lazy',
-        isa     => InstanceOf ["Sereal::$_"],
-        default => qsub "Sereal::$_->new",
-    );
-}
-
 #pod =head1 HANDLERS
 #pod
 #pod This module works by adding C<request_send> and C<response_done>
@@ -163,64 +158,100 @@ for (qw(Encoder Decoder)) {
 sub BUILD {
     my $self = shift;
 
-    $self->add_handler(    # load from cache on each GET request
-        request_send => sub {
-            my $request = shift;
-            $self->_set_is_cached(0);
-            if ( not $self->ref_in_cache_key ) {
-                my $clone = $request->clone;
-                $clone->header( Referer => undef );
-                $request = $clone->as_string;
-            }
-
-            my $response = $self->cache->get($request);
-            $response
-                &&= sereal_decode_with_object( $self->_decoder, $response );
-
-            return
-                   if not($response)
-                or $response->code < HTTP_OK
-                or $response->code > HTTP_MOVED_PERMANENTLY;
-            $self->_set_is_cached(1);
-            return $response;
-        },
-        ( m_method => 'GET' ),
-    );
-
-    $self->add_handler(    # save to cache after successful GET
-        response_done => sub {
-            return if not my $response = shift;
-
-            if (not( $response->header('client-transfer-encoding')
-                    and 'ARRAY' eq
-                    ref $response->header('client-transfer-encoding')
-                    and any { 'chunked' eq $_ }
-                    @{ $response->header('client-transfer-encoding') } )
-                )
-            {
-                for ( $response->header('size') ) {
-                    return
-                        if not defined and $self->cache_undef_content_length;
-                    return
-                        if 0 == $_
-                        and not $self->cache_zero_content_length;
-                    return
-                        if $_ != length $response->content
-                        and not $self->cache_mismatch_content_length;
-                }
-            }
-
-            $response->decode;
-            $self->cache->set( $response->request->as_string =>
-                    sereal_encode_with_object( $self->_encoder, $response ) );
-            return;
-        },
+    $self->add_handler( request_send => \&_get_cache, ( m_method => 'GET' ) );
+    $self->add_handler(
+        response_done => \&_set_cache,
         ( m_method => 'GET', m_code => 2 ),
     );
+    $self->add_handler(
+        response_header => \&_get_not_modified,
+        ( m_method => 'GET', m_code => HTTP_NOT_MODIFIED ),
+    );
 
     return;
 }
 
+# load from cache on each GET request
+sub _get_cache {
+    my ( $request, $self ) = @_;
+    $self->_set_is_cached(0);
+
+    my $clone = $request->clone;
+    if ( not $self->ref_in_cache_key ) { $clone->header( Referer => undef ) }
+    return if $self->_no_cache_header_directives($request);
+
+    return if not my $response = $self->cache->get( $clone->as_string );
+    return
+        if $response->code < HTTP_OK
+        or $response->code > HTTP_MOVED_PERMANENTLY;
+
+    if ( $response->header('etag') ) {
+        $clone->header( if_none_match => $response->header('etag') );
+        $response = $self->request($clone);
+    }
+    return if $self->_no_cache_header_directives($response);
+
+    $self->_set_is_cached(1);
+    return $response;
+}
+
+sub _get_not_modified {
+    my ( $response, $self ) = @_;
+    $self->_set_is_cached(0);
+
+    my $request = $response->request->clone;
+    $request->remove_header(qw(if_modified_since if_none_match));
+
+    my $cached_response = $self->cache->get( $request->as_string );
+    $response->content( $cached_response->decoded_content );
+
+    $self->_set_is_cached(1);
+    return;
+}
+
+# save to cache after successful GET
+sub _set_cache {
+    my ( $response, $self ) = @_;
+    return if not $response;
+
+    if (not(    $response->header('client-transfer-encoding')
+            and 'ARRAY' eq ref $response->header('client-transfer-encoding')
+            and any { 'chunked' eq $_ }
+            @{ $response->header('client-transfer-encoding') } )
+        )
+    {
+        for ( $response->header('size') ) {
+            return
+                if not defined and $self->cache_undef_content_length;
+            return
+                if 0 == $_
+                and not $self->cache_zero_content_length;
+            return
+                if $_ != length $response->content
+                and not $self->cache_mismatch_content_length;
+        }
+    }
+
+    for my $message ( $response, $response->request ) {
+        return if $self->_no_cache_header_directives($message);
+    }
+
+    $response->decode;
+    $response->remove_content_headers;
+    $self->cache->set( $response->request->as_string => $response );
+    return;
+}
+
+sub _no_cache_header_directives {
+    my ( $self, $message ) = @_;
+    for my $header_name (qw(pragma cache_control)) {
+        if ( my @directives = $message->header($header_name) ) {
+            return 1 if any {/\A no- (?: cache | store ) /xms} @directives;
+        }
+    }
+    return;
+}
+
 #pod =for Pod::Coverage FOREIGNBUILDARGS
 #pod
 #pod =cut
@@ -248,7 +279,7 @@ LWPx::UserAgent::Cached - Subclass of LWP::UserAgent that caches HTTP GET reques
 
 =head1 VERSION
 
-version 0.004
+version 0.006
 
 =head1 SYNOPSIS
 
@@ -270,6 +301,13 @@ inheriting from L<WWW::Mechanize|WWW::Mechanize>;
 instead it is just a direct subclass of
 L<LWP::UserAgent|LWP::UserAgent>.
 
+=head2 HTTP/1.1 cache operation
+
+Full HTTP/1.1 cache compliance is a work in progress. As of version 0.006 we
+have limited support for HTTP/1.1 C<ETag>/C<If-None-Match> headers, as well as
+C<no-cache> and C<no-store> C<Cache-Control> directives (both on request and
+response) and the C<Pragma: no-cache> request header.
+
 =head1 ATTRIBUTES
 
 =head2 cache
